import numpy as np
import pandas as pd
from .restart_pdb_to_df import *

def calc_rmsError(PathName: str, fitted_sphere_params: list):
    
    # positions = fake_PDB_pdb_to_df("show_structure.pdb")
    positions = fake_PDB_pdb_to_df(PathName)
    # convert coordinate unit from angstrom to nm
    positions["x_coord"] = positions["x_coord"]/10.0
    positions["y_coord"] = positions["y_coord"]/10.0
    positions["z_coord"] = positions["z_coord"]/10.0

    # get the number of monomers
    monomer_count = 0
    for i in range(len(positions)):
        if(positions.iloc[i]["Cite_Name"] == "COM"):
            monomer_count += 1
    
    # get count of interfaces for each monomer
    interfaces_count = []
    count = 0
    for i in range(len(positions)):
        count += 1
        if(positions.iloc[i]["Cite_Name"] == "COM"):
            count = 0
        if(i+1 == len(positions) or positions.iloc[i+1]["Cite_Name"] == "COM"):
            interfaces_count.append(count)
    interfaces_count = np.array(interfaces_count)

    # get the index of the COM
    COM_index = []
    curr_index = 0
    for i in range(len(interfaces_count)):
        COM_index.append(curr_index)
        curr_index += (interfaces_count[i]+1)

    # get the distance from the sphere center
    distance_r = []
    for i in range(len(COM_index)):
        COM_coord = [positions.iloc[COM_index[i]]["x_coord"], positions.iloc[COM_index[i]]["y_coord"], positions.iloc[COM_index[i]]["z_coord"]]
        distance_from_sphere_center = np.linalg.norm(np.array(COM_coord) - np.array(fitted_sphere_params[:3]))
        distance_r.append(distance_from_sphere_center)
    distance_r = np.array(distance_r)
    distance_r_std = np.std(distance_r)
    # print(distance_r)
    print("std from sphere center: ", distance_r_std)
    return 0
    





import numpy as np
import pandas as pd
from .restart_pdb_to_df import *

def calc_fit_sphere(PathName: str):

    positions = fake_PDB_pdb_to_df(PathName)

    # convert coordinate unit from angstrom to nm
    positions["x_coord"] = positions["x_coord"]/10.0
    positions["y_coord"] = positions["y_coord"]/10.0
    positions["z_coord"] = positions["z_coord"]/10.0
    COM_positions = positions[positions["Cite_Name"] == "COM"]
    res = []
    for i in range(int((len(COM_positions)/50) **4)):
        # sample four random COMs
        
        rand1 = int(np.random.random()* len(COM_positions))
        rand2 = int(np.random.random()* len(COM_positions))
        while rand1 == rand2:
            rand2 = int(np.random.random()* len(COM_positions))
        rand3 = int(np.random.random()* len(COM_positions))
        while rand3 == rand2 or rand3 == rand1:
            rand3 = int(np.random.random()* len(COM_positions))
        rand4 = int(np.random.random()* len(COM_positions))
        while rand4 == rand3 or rand4 == rand2 or rand4 == rand1:
            rand4 = int(np.random.random()* len(COM_positions))
        
        

        sampled_coordinates = np.zeros([4,3])
        
        sampled_coordinates[0] = np.array([COM_positions.iloc[rand1]["x_coord"], COM_positions.iloc[rand1]["y_coord"], COM_positions.iloc[rand1]["z_coord"]])
        sampled_coordinates[1] = np.array([COM_positions.iloc[rand2]["x_coord"], COM_positions.iloc[rand2]["y_coord"], COM_positions.iloc[rand2]["z_coord"]])
        sampled_coordinates[2] = np.array([COM_positions.iloc[rand3]["x_coord"], COM_positions.iloc[rand3]["y_coord"], COM_positions.iloc[rand3]["z_coord"]])
        sampled_coordinates[3] = np.array([COM_positions.iloc[rand4]["x_coord"], COM_positions.iloc[rand4]["y_coord"], COM_positions.iloc[rand4]["z_coord"]])

        A = 2 * (sampled_coordinates[:-1] - sampled_coordinates[-1])

        # Constant vector B
        B = np.sum(sampled_coordinates[:-1]**2, axis=1) - np.sum(sampled_coordinates[-1]**2)

        # Solve the system of equations Ax = B
        center = np.linalg.solve(A, B)

        # Calculate the radius
        radius = np.linalg.norm(sampled_coordinates[0] - center)
        
        res.append(np.array([COM_positions.iloc[rand1]["Protein_Num"], 
                                                   COM_positions.iloc[rand2]["Protein_Num"], 
                                                   COM_positions.iloc[rand3]["Protein_Num"], center[0], center[1], center[2], radius]))
    res = np.array(res)
    pd.DataFrame(res).to_csv(PathName[:-4]+"_varience.csv", index=False, header=False)
    print(np.mean(res[:, 3:], axis=0))
    return np.mean(res[:, 3:], axis=0)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from .restart_pdb_to_df import *

def plot_3D_sites(positionsVec, COM_index, from_PDB = False, PathName = "",chains_included = []):
    fig = plt.figure(1)
    color_list = ['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9']
    ax = plt.axes(projection="3d")
    if from_PDB:
        if PathName == "":
            raise Exception("Please provide the path to the PDB file")
        
        positions = fake_PDB_pdb_to_df(PathName)
        # convert coordinate unit from angstrom to nm
        positions["x_coord"] = positions["x_coord"]/10.0
        positions["y_coord"] = positions["y_coord"]/10.0
        positions["z_coord"] = positions["z_coord"]/10.0
        
        if len(chains_included) != 0:
            positions = positions[positions["Protein_Name"].isin(chains_included)]
        # get the number of monomers
        monomer_count = 0
        for i in range(len(positions)):
            if(positions.iloc[i]["Site_Name"] == "COM"):
                monomer_count += 1
        
        # get count of interfaces for each monomer
        interfaces_count = []
        count = 0
        for i in range(len(positions)):
            count += 1
            if(positions.iloc[i]["Site_Name"] == "COM"):
                count = 0
            if(i+1 == len(positions) or positions.iloc[i+1]["Site_Name"] == "COM"):
                interfaces_count.append(count)
        interfaces_count = np.array(interfaces_count)

        # get the index of the COM
        COM_index = []
        curr_index = 0
        for i in range(len(interfaces_count)):
            COM_index.append(curr_index)
            curr_index += (interfaces_count[i]+1)

        # get the coordinates of the monomers
        positionsVec = np.zeros([len(positions),3])
        for i in range(len(positions)):
            positionsVec[i,:] = [positions.iloc[i]["x_coord"], positions.iloc[i]["y_coord"], positions.iloc[i]["z_coord"]]
        


    for i in range(len(COM_index)):
        ax.scatter(positionsVec[COM_index[i]][0], positionsVec[COM_index[i]][1],positionsVec[COM_index[i]][2], color=color_list[i % 9])
        if from_PDB:
            ax.text(positionsVec[COM_index[i]][0], positionsVec[COM_index[i]][1],
                    positionsVec[COM_index[i]][2], positions.iloc[COM_index[i]]["Protein_Name"], color='k')
        next_COM_index = COM_index[i+1] if i+1 < len(COM_index) else len(positionsVec)
        for j in range(COM_index[i]+1, next_COM_index):
            figure = ax.plot([positionsVec[COM_index[i]][0], positionsVec[j][0]],
                             [positionsVec[COM_index[i]][1], positionsVec[j][1]],
                             [positionsVec[COM_index[i]][2], positionsVec[j][2]], color=color_list[i % 9])
    ax.set_xlabel('x (nm)')
    ax.set_ylabel('y (nm)')
    ax.set_zlabel('z (nm)')
    plt.show()
    return 0


import numpy as np

def find_complementry_site(p1,positions):
    target_site_name = ""
    p1_index = -1

    # figure out the target site name
    for i in range(len(positions)):
        if positions.iloc[i]["x_coord"] == p1[0] and positions.iloc[i]["y_coord"] == p1[1] and positions.iloc[i]["z_coord"] == p1[2]:
            target_site_name = positions.iloc[i]["Site_Name"]
            p1_index = i
            break
    # find the complementry site to the target site
    site_pair = positions[positions["Site_Name"] == target_site_name]
    if len(site_pair) != 2:
        raise Exception("The target site does not have a pair or has more than one pair.")
    complementary_site = site_pair[site_pair.index != p1_index]
    complementary_site_index = complementary_site.index
    p2 = np.array([complementary_site.iloc[0]["x_coord"],
                    complementary_site.iloc[0]["y_coord"],
                      complementary_site.iloc[0]["z_coord"]])
    # find the COM of the monomer with the complementary site
    complementary_site_COM_index = -1
    for i in range(len(positions)):
        if positions.iloc[i]["Site_Name"] == "COM":
            complementary_site_COM_index = i
        if i >= complementary_site_index:
            break
    complementary_site_COM = positions.iloc[complementary_site_COM_index]
    c2 = np.array([complementary_site_COM["x_coord"], complementary_site_COM["y_coord"], complementary_site_COM["z_coord"]])
    return c2, p2
    


import numpy as np

def calculateAngles(c1, c2, p1, p2, n1, n2):
        """
        Determine the angles of the reaction (theta1, theta2, phi1, phi2, omega) given the coordinates of the two Center of Mass (c1 and c2) and two reaction sites (p1 and p2), and two norm vectors (n1 and n2).

        Parameters
        ----------
        c1 : numpy.array
            Center of Mass vector for the first molecule.
        c2 : numpy.array
            Center of Mass vector for the second molecule.
        p1 : numpy.array
            Reaction site vector for the first molecule.
        p2 : numpy.array
            Reaction site vector for the second molecule.
        n1 : numpy.array
            Norm vector for the first molecule.
        n2 : numpy.array
            Norm vector for the second molecule.

        Returns
        -------
        tuple
            The tuple (theta1, theta2, phi1, phi2, omega), where theta1, theta2, phi1, phi2, omega are the angles in radians.
        """
        v1 = p1 - c1
        v2 = p2 - c2
        sigma1 = p1 - p2
        sigma2 = -sigma1

        if np.dot(v1, sigma1) / (np.linalg.norm(v1) * np.linalg.norm(sigma1)) <= 1:  # prevent float point arithematic rounding errors
            theta1 = np.arccos(
                np.dot(v1, sigma1) / (np.linalg.norm(v1) * np.linalg.norm(sigma1))
            )
        else:
            theta1 = 0


        if np.dot(v2, sigma2) / (np.linalg.norm(v2) * np.linalg.norm(sigma2)) <= 1:   # prevent float point arithematic rounding errors
            theta2 = np.arccos(
                np.dot(v2, sigma2) / (np.linalg.norm(v2) * np.linalg.norm(sigma2))
            )
        else:
            theta2 = 0

        t1 = np.cross(v1, sigma1)
        t2 = np.cross(v1, n1)
        norm_t1 = t1 / np.linalg.norm(t1)
        norm_t2 = t2 / np.linalg.norm(t2)
        if np.dot(norm_t1, norm_t2) <= 1:  # prevent float point arithematic rounding errors
            phi1 = np.arccos(np.dot(norm_t1, norm_t2))
        else:
            phi1 = 0

        # the sign of phi1 is determined by the direction of t2 relative to the right-hand rule of cross product of v1 and t1
        if np.dot(np.cross(v1, t1), t2) > 0:
            phi1 = -phi1

        t1 = np.cross(v2, sigma2)
        t2 = np.cross(v2, n2)
        norm_t1 = t1 / np.linalg.norm(t1)
        norm_t2 = t2 / np.linalg.norm(t2)
        if np.dot(norm_t1, norm_t2) <= 1:   # prevent float point arithematic rounding errors
            phi2 = np.arccos(np.dot(norm_t1, norm_t2))
        else:  
            phi2 = 0

        # the sign of phi2 is determined by the direction of t2 relative to the right-hand rule of cross product of v2 and t1
        if np.dot(np.cross(v2, t1), t2) > 0:
            phi2 = -phi2

        if not np.isclose(np.linalg.norm(np.cross(v1, sigma1)), 0) and not np.isclose(
            np.linalg.norm(np.cross(v2, sigma2)), 0
        ):
            t1 = np.cross(sigma1, v1)
            t2 = np.cross(sigma1, v2)
        else:
            t1 = np.cross(sigma1, n1)
            t2 = np.cross(sigma1, n2)
        

        if np.dot(t1, t2) / (np.linalg.norm(t1) * np.linalg.norm(t2)) <= 1:  # prevent float point arithematic rounding errors
            omega = np.arccos(np.dot(t1, t2) / (np.linalg.norm(t1) * np.linalg.norm(t2)))
        else:
            omega = 0
        if np.abs(omega) < 1.4902e-8:
            omega = 0

        # the sign of omega is determined by the direction of t2 relative to the right-hand rule of cross product of sigma1 and t1
        if np.dot(np.cross(sigma1, t1), t2) > 0:
            omega = -omega

        return np.linalg.norm(sigma1), theta1, theta2, phi1, phi2, omega


import numpy as np
import pandas as pd
from .calculate_rmsd import *
from .calculate_gradient import *
from .determine_gagTemplate_structure import *
from .translate_gags_on_sphere import *
from .xyz_to_sphere_coordinates import *
from .restart_pdb_to_df import *
from .gag_write_PDB import *
from ..database_PDB.dtb_PDB_separate_read import *
from ..database_PDB.dtb_PDB_write_PDB import *
from .plot_3D_sites import *
from .fake_calc_angle import *
from .find_complementry_site import *


def repeated_protein_subunit_regularization(PathName: str, UniqueChainList: list = [[]]):
    if(type(UniqueChainList[0]) != list):
        raise TypeError("UniqueChainList must be a 2 dimensional list")
    
    # positions = fake_PDB_pdb_to_df("show_structure.pdb")
    positions = fake_PDB_pdb_to_df(PathName)
    # convert coordinate unit from angstrom to nm
    positions["x_coord"] = positions["x_coord"]/10.0
    positions["y_coord"] = positions["y_coord"]/10.0
    positions["z_coord"] = positions["z_coord"]/10.0

    if UniqueChainList == [[]]:
        UniqueChainList = [positions["Protein_Name"].unique()]
    
    # get the number of monomers
    monomer_count = 0
    for i in range(len(positions)):
        if(positions.iloc[i]["Site_Name"] == "COM"):
            monomer_count += 1

    # get count of interfaces for each monomer
    interfaces_count = []
    count = 0
    for i in range(len(positions)):
        count += 1
        if(positions.iloc[i]["Site_Name"] == "COM"):
            count = 0
        if(i+1 == len(positions) or positions.iloc[i+1]["Site_Name"] == "COM"):
            interfaces_count.append(count)
    interfaces_count = np.array(interfaces_count)

    # get the index of the COM
    COM_index = []
    curr_index = 0
    for i in range(len(interfaces_count)):
        COM_index.append(curr_index)
        curr_index += (interfaces_count[i]+1)


    # sort the sites in the order of increasing distance from the COM of each monomer.
    distances = []
    for i in range(monomer_count):
        COM_coord = [positions.iloc[COM_index[i]]["x_coord"], positions.iloc[COM_index[i]]["y_coord"], positions.iloc[COM_index[i]]["z_coord"]]
        distances.append(0)
        for j in range(interfaces_count[i]):
            interface_coord = [positions.iloc[COM_index[i]+j+1]["x_coord"], positions.iloc[COM_index[i]+j+1]["y_coord"], positions.iloc[COM_index[i]+j+1]["z_coord"]]
            distances.append(np.linalg.norm(np.array(COM_coord)-np.array(interface_coord)))
    positions["Distance"] = distances

    for i in range(monomer_count):
        for j in range(COM_index[i], COM_index[i+1] if i+1 < len(COM_index) else len(positions)):
            for k in range(j, COM_index[i+1] if i+1 < len(COM_index) else len(positions)):
                if(positions.iloc[j]["Distance"] > positions.iloc[k]["Distance"]):
                    temp = positions.iloc[j]
                    positions.iloc[j] = positions.iloc[k]
                    positions.iloc[k] = temp
    
    # record the positions of the COM and interfaces of each monomer in a np array
    positionsVec = np.zeros([len(positions),3])
    for i in range(len(positions)):
        positionsVec[i,:] = [positions.iloc[i]["x_coord"], positions.iloc[i]["y_coord"], positions.iloc[i]["z_coord"]]

    # plot before reshaping
    plot_3D_sites(positionsVec, COM_index)

    # get the coordinates of the COM
    centersVec = np.zeros([monomer_count,3])
    for i in range(len(COM_index)):
        centersVec[i] = positionsVec[COM_index[i]]
    ##############################################
    # find the sphere radius and the sphere center
    ##############################################
    sphereXYZR = [0,0,0,70] # initial trial values for sphere x, y, z, and R respectively 
    rmsdOld = calculate_rmsd(centersVec,sphereXYZR)
    isForceSmallEnough = False
    while isForceSmallEnough == False :
        force = calculate_gradient(centersVec,sphereXYZR)
        stepSize = 1.0
        tempXYZR = sphereXYZR - force * stepSize
        rmsdNew = calculate_rmsd(centersVec,tempXYZR)
        while rmsdNew > rmsdOld :
            stepSize = stepSize * 0.8
            tempXYZR = sphereXYZR - force * stepSize
            rmsdNew = calculate_rmsd(centersVec,tempXYZR)
        sphereXYZR = tempXYZR
        rmsdOld = calculate_rmsd(centersVec,sphereXYZR)
        if ( np.linalg.norm(force) < 0.01 ):
            isForceSmallEnough = True

    print('Sphere center position [x,y,z] and radius [R] are, respectively: \n',sphereXYZR) 
    print("\n")
    x0 = sphereXYZR[0] 
    y0 = sphereXYZR[1] 
    z0 = sphereXYZR[2] 
    r0 = sphereXYZR[3] 

    ##############################################
    # Second, move the center of the sphere to the axis origin, equivalently move the monomers 
    positionsVec[:,0] = positionsVec[:,0] - x0
    positionsVec[:,1] = positionsVec[:,1] - y0
    positionsVec[:,2] = positionsVec[:,2] - z0
    centersVec[:,0] = centersVec[:,0] - x0
    centersVec[:,1] = centersVec[:,1] - y0
    centersVec[:,2] = centersVec[:,2] - z0

    ##############################################
    # Third, move the centers of monomers to the sphere surface
    for i in range (0,monomer_count):
        center = centersVec[i,:]
        move = center/np.linalg.norm(center) * r0 - center
        centersVec[i,:] = centersVec[i,:] + move
        for j in range (0,interfaces_count[i]):
            positionsVec[COM_index[i] + j + 1,:] = positionsVec[COM_index[i] + j + 1,:] + move
    
    # plot after fitting to sphere
    plot_3D_sites(positionsVec, COM_index)


    # undate positions dataframe
    for i in range(len(positions)):
        positions.at[i,"x_coord"] = positionsVec[i,0]
        positions.at[i,"y_coord"] = positionsVec[i,1]
        positions.at[i,"z_coord"] = positionsVec[i,2]

    
    ##############################################
    # For each group of unqiue chains,
    # Determine the regulation coefficients and apply the regulation to the monomers
    ##############################################
    # find the "full monomers"
    # monomers that have all interfaces recorded in the PDB file are considered as full monomers
    for unique_chains in UniqueChainList:
        # get the indecies of the COM for this group of unique chains
        unique_chains_COM_index = []
        for i in range(len(positions)):
            if positions.iloc[i]["Protein_Name"] in unique_chains and positions.iloc[i]["Site_Name"] == "COM":
                unique_chains_COM_index.append(i)
        # get the interfaces count for this group of unique chains
        unique_chains_interfaces_count = []
        for i in range(len(COM_index)):
            if COM_index[i] in unique_chains_COM_index:
                unique_chains_interfaces_count.append(interfaces_count[i])
        # get the indecies of the COM for each full monomer in this group of unique chains (monomers that have all interfaces recorded in the PDB file)
        unique_chains_full_monomer_COM_index = []
        unique_chains_interfaces_count = np.array(unique_chains_interfaces_count)
        unique_chains_full_interfaces_count = np.max(unique_chains_interfaces_count)
        for i in range(len(unique_chains_interfaces_count)):
            if(unique_chains_interfaces_count[i] == unique_chains_full_interfaces_count):
                unique_chains_full_monomer_COM_index.append(unique_chains_COM_index[i])
        unique_chains_full_monomer_count = len(unique_chains_full_monomer_COM_index)

        # generate a np array to record the coordinates of the COM and interfaces of the full monomers
        full_monomer_positionsVec = np.zeros([unique_chains_full_monomer_count*(unique_chains_full_interfaces_count+1),3])
        for i in range(unique_chains_full_monomer_count):
            for j in range(unique_chains_full_interfaces_count+1):
                full_monomer_positionsVec[i*(unique_chains_full_interfaces_count+1)+j] = positionsVec[unique_chains_full_monomer_COM_index[i]+j,:]
        

        full_monomer_COM_index_graph = []
        for i in range(unique_chains_full_monomer_count):
            full_monomer_COM_index_graph.append(i*(unique_chains_full_interfaces_count+1))
        
        # plot this group of unqiue chains before regularization
        plot_3D_sites(full_monomer_positionsVec, full_monomer_COM_index_graph)

        # Determine the regularized coefficients. All monomers will be reshaped according to this template
        # monomerTemplate is the positions of the gag center and five interfaces
        # monomerTemplateInterCoeffs is the coefficients of the gag 5 interfaces in the internal basis system
        numSites = unique_chains_full_interfaces_count + 1
        monomerCoeff = determine_gagTemplate_structure(unique_chains_full_monomer_count, numSites, full_monomer_positionsVec, returnCoeff = True)
        print("Chains:", unique_chains)
        print("Interfaces at: [x, y, z]")
        print(monomerCoeff)
        print("\n")
        

        # set up the internal coordinate system of the monomers: 3 basis vecs: interBaseVec0, interBaseVec1, interBaseVec2
        # and apply regularization coefficients
        regularized_positionsVec = np.zeros([unique_chains_full_monomer_count*(numSites),3])
        deviation = np.zeros([unique_chains_full_monomer_count*(numSites)])
        for i in range(0,unique_chains_full_monomer_count):
            center = full_monomer_positionsVec[numSites*i,:]                                            # center of the monomer
            interBaseVec0 = center / np.linalg.norm(center)                   # along the radius direction
            interBaseVec1 = full_monomer_positionsVec[i*numSites+1,:] - center          # in the direction of the first interface
            interBaseVec2 = np.cross(interBaseVec0,interBaseVec1)               # orthogonal to the first two basis vecs
            interBaseVec2 = interBaseVec2 / np.linalg.norm(interBaseVec2) 
            interBaseVec1 = np.cross(interBaseVec2,interBaseVec0)
            interBaseVec1 = interBaseVec1 / np.linalg.norm(interBaseVec1)
            regularized_positionsVec[numSites*i,:] = center
            for j in range (0,numSites-1) :
                regularized_positionsVec[numSites*i+j+1,:] = center + interBaseVec0 * monomerCoeff[j,0] + interBaseVec1 * monomerCoeff[j,1] + interBaseVec2 * monomerCoeff[j,2]
                original_interface_position = full_monomer_positionsVec[numSites*i+j+1,:]
                deviation[numSites*i+j+1] = np.linalg.norm(regularized_positionsVec[numSites*i+j+1] - original_interface_position)
        # print("Deviation = ", deviation)


        # plot this group of unqiue chains after regularization
        plot_3D_sites(regularized_positionsVec, full_monomer_COM_index_graph)

        # update the coordinates of the full monomers
        for i in range(unique_chains_full_monomer_count):
            for j in range(unique_chains_full_interfaces_count+1):
                positions.at[unique_chains_full_monomer_COM_index[i]+j, "x_coord"] = regularized_positionsVec[i*(unique_chains_full_interfaces_count+1)+j,0]
                positions.at[unique_chains_full_monomer_COM_index[i]+j, "y_coord"] = regularized_positionsVec[i*(unique_chains_full_interfaces_count+1)+j,1]
                positions.at[unique_chains_full_monomer_COM_index[i]+j, "z_coord"] = regularized_positionsVec[i*(unique_chains_full_interfaces_count+1)+j,2]
                positionsVec[unique_chains_full_monomer_COM_index[i]+j,:] = regularized_positionsVec[i*(unique_chains_full_interfaces_count+1)+j,:]
    
    # plot after regularization
    plot_3D_sites(positionsVec, COM_index)
    
    for unique_chains in UniqueChainList:
        print("With normal vector (1,0,0),")
        print("The binding parameters for the interfaces in chains", unique_chains, "are:")
        print(" [ sigma,      theta1,     theta2,     phi1,       phi2,       omega]")
        # get the indecies of the COM for this group of unique chains
        unique_chains_COM_index = []
        for i in range(len(positions)):
            if positions.iloc[i]["Protein_Name"] in unique_chains and positions.iloc[i]["Site_Name"] == "COM":
                unique_chains_COM_index.append(i)
        
        # get the interfaces count for this group of unique chains
        unique_chains_interfaces_count = []
        for i in range(len(COM_index)):
            if COM_index[i] in unique_chains_COM_index:
                unique_chains_interfaces_count.append(interfaces_count[i])

        # get the indecies of the COM for each full monomer in this group of unique chains (monomers that have all interfaces recorded in the PDB file)
        unique_chains_full_monomer_COM_index = []
        unique_chains_interfaces_count = np.array(unique_chains_interfaces_count)
        unique_chains_full_interfaces_count = np.max(unique_chains_interfaces_count)
        for i in range(len(unique_chains_interfaces_count)):
            if(unique_chains_interfaces_count[i] == unique_chains_full_interfaces_count):
                unique_chains_full_monomer_COM_index.append(unique_chains_COM_index[i])
        unique_chains_full_monomer_count = len(unique_chains_full_monomer_COM_index)
        
        # calculate the sigma and angles of reaction for each chain in this group of unique chains
        angles_list = []
        for i in unique_chains_full_monomer_COM_index:
            c1 = positionsVec[i]
            for j in range(unique_chains_full_interfaces_count):
                p1 = positionsVec[i+j+1,:]
                c2, p2 = find_complementry_site(p1,positions)
                n1 = c1
                n2 = c2
                angles_list.append(np.array(calculateAngles(c1,c2,p1,p2,n1,n2)))
        angles_list = np.array(angles_list)
        # for i in angles_list:
        #     print(i)

        final_mean_angle_list = []
        # take the average angles and sigma values 
        for i in range(unique_chains_full_interfaces_count):
            temp_angles_list = []
            mean_angles_list = []
            for j in range(len(unique_chains_full_monomer_COM_index)):
                temp_angles_list.append(angles_list[j*unique_chains_full_interfaces_count + i])
            temp_angles_list = np.array(temp_angles_list)
            mean_angles_list = np.mean(temp_angles_list, axis = 0)
            final_mean_angle_list.append(mean_angles_list)
        final_mean_angle_array = np.array(final_mean_angle_list)
        print(final_mean_angle_array)    
        print("\n")

    return positionsVec
